{-# LANGUAGE TemplateHaskell #-}

module ISets where

import qualified Data.List       as L
import           Test.QuickCheck

newtype ISet = ISet [(Int,Int)]
  deriving (Eq,Show)

valid :: ISet -> Bool
valid (ISet xys) =
  all (uncurry (<=)) xys &&
  and [y+1<x | ((_,y),(x,_)) <- zip xys (drop 1 xys)]

-- toList --------------------------------------------------------------

toList :: ISet -> [Int]
toList (ISet xys) =
  concat [[x..y] | (x,y) <- xys]

-- Generator for ISets -------------------------------------------------

instance Arbitrary ISet where
  arbitrary = do Ordered xs <- arbitrary
                 return $ ISet $ make xs
    where make (x:y:xs) = (x,y):make (map (+2) xs)
          make _        = []
  shrink (ISet xs) = filter valid $ map ISet $ shrink xs

-- member --------------------------------------------------------------

prop_Member x s =
  (x `member` s) === (x `elem` toList s)

member x (ISet xys) = member' x xys

member' x [] = False
member' x ((y,z):yzs)
  | x > z = member' x yzs
member' x ((y,z):yzs) =
  x >= y && x <= z
  
-- insert --------------------------------------------------------------

prop_InsertValid x s = valid (insert x s)

prop_InsertModel x s =
  toList (insert x s) === L.insert x (toList s L.\\ [x])

insert :: Int -> ISet -> ISet
insert x (ISet xys) = ISet (insert' x xys)

insert' x [] = [(x,x)]
insert' x ((a,b):(c,d):s)
  | b+1==x && x+1==c = (a,d):s
insert' x ((y,z):yzs)
  | x>=y && x<=z = (y,z):yzs
  | x==z+1       = (y,x):yzs
  | x>z+1        = (y,z):insert' x yzs
  | x+1<y        = (x,x):(y,z):yzs
  | x+1==y       = (x,z):yzs

-- delete --------------------------------------------------------------

prop_DeleteValid x s = valid (delete x s)

prop_DeleteModel x s =
  toList (delete x s) === L.delete x (toList s)

delete x (ISet xys) = ISet (delete' x xys)

delete' x [] = []
delete' x ((y,z):yzs)
  | x<y	         = (y,z):yzs
  | x>z          = (y,z):delete' x yzs
  | x==y && x==z = yzs
  | x==y         = (y+1,z):yzs
  | x==z         = (y,z-1):yzs
  | otherwise    = (y,x-1):(x+1,z):yzs

-- union ---------------------------------------------------------------

prop_UnionValid s s' = valid (union s s')

prop_UnionModel s s' =
  toList (union s s') === L.sort (L.union (toList s) (toList s'))

union (ISet xys) (ISet xys') = ISet (union' xys xys')

union' [] ys = ys
union' xs [] = xs
union' ((a,b):xs) ((c,d):ys)
  | b+1<c = (a,b):union' xs ((c,d):ys)
  | d+1<a = (c,d):union' ((a,b):xs) ys
  | b<=d  = union' xs ((min a c,d):ys)
  | d<=b  = union' ((min a c,b):xs) ys

-- Enable running all properties ---------------------------------------
return []

runTests :: IO Bool
runTests = $quickCheckAll
